/*	PCKBRD3 - A PC keyboard interface program for use with the Bagotronix
	DOS Stamp(tm).

	Copyright 2001 Bagotronix Inc.
	Authors: D. P. Knapp, Ivan Baggett
	Jan. 31, 2001:  first release

	This should be considered experimental software.  No express or implied
	warranties exist.  Use and modify for your needs.

	This program requires an AT or PS/2 style keyboard connected as follows:
	DOS Stamp          AT Keyboard
	PIO31 (JP1.17)     DATA (pin 2)
	PIO0 (JP1.21)      CLOCK (pin 1)
	VCC (JP1.33)       +5V (pin 5)
	GND (JP1.34)       GND (pin 4)
*/

#include <dos.h>
#include <stdio.h>
#include <stdlib.h>
#include <ctype.h>
#include "stdinc.h"
#include "dosstamp.h"

UBYTE last_scan_code = 0;

static UBYTE at_kbrd_status;
static UBYTE at_ascii_code;
static UBYTE at_scan_code;
static UBYTE at_leds_status;
static BOOLEAN numpad_flag;

unsigned char LOOKUP [132] [4]= {

/////////-----------------------SCAN    KEY    NORMAL SHIFT  ALT    CTRL-----

	 {0   ,0   ,0   ,0   }, //00   //////////////////////////////////
	 {0x43,0x5c,0x70,0x66}, //01    F9     0x43   0x5c   0x70   0x66
	 {0   ,0   ,0   ,0   }, //02   //////////////////////////////////
	 {0x3f,0x58,0x6c,0x62}, //03    F5     0x3f   0x58   0x6c   0x62
	 {0x3d,0x56,0x6a,0x60}, //04    F3     0x3d   0x56   0x6a   0x60
	 {0x3b,0x54,0x68,0x5e}, //05    F1     0x3b   0x54   0x68   0x5e
	 {0x3c,0x55,0x69,0x5f}, //06    F2     0x3c   0x55   0x69   0x5f
	 {0x86,0x88,0x8c,0x8a}, //07    F12    0x86   0x88   0x8c   0x8a
	 {0   ,0   ,0   ,0   }, //08   //////////////////////////////////
	 {0x44,0x5d,0x71,0x67}, //09    F10    0x44   0x5d   0x71   0x67
	 {0x42,0x5b,0x6f,0x65}, //0a    F8     0x42   0x5b   0x6f   0x65
	 {0x40,0x59,0x6d,0x63}, //0b    F6     0x40   0x59   0x6d   0x63
	 {0x3e,0x57,0x6b,0x61}, //0c    F4     0x3e   0x57   0x6b   0x61
	 {9   ,0x0f,0xa5,0x94}, //0d    TAB    9      0x0f   0xa5   0x94
	 {'`' ,'~' ,0x29,0   }, //0e    ` ~    (`)    (~)    0x29   0
	 {0   ,0   ,0   ,0   }, //0f   //////////////////////////////////
	 {0   ,0   ,0   ,0   }, //10   //////////////////////////////////

/////////-----------------------SCAN    KEY    NORMAL SHIFT  ALT    CTRL-----

	 {0   ,0   ,0   ,0   }, //11    ALT    ---------status----------
	 {0   ,0   ,0   ,0   }, //12    SHIFT  ---------status----------
	 {0   ,0   ,0   ,0   }, //13   //////////////////////////////////
	 {0   ,0   ,0   ,0   }, //14    CTRL   ---------status----------
	 {'q' ,'Q' ,0x10,0x11}, //15    q Q    (q)    (Q)    0x10   0x11
	 {'1' ,'!' ,0x78,0   }, //16    1 !    (1)    (!)    0x78   0
	 {0   ,0   ,0   ,0   }, //17   //////////////////////////////////
	 {0   ,0   ,0   ,0   }, //18   //////////////////////////////////
	 {0   ,0   ,0   ,0   }, //19   //////////////////////////////////
	 {'z' ,'Z' ,0x2c,0x1a}, //1a    z Z    (z)    (Z)    0x2c   0x1a
	 {'s' ,'S' ,0x1f,0x13}, //1b    s S    (s)    (S)    0x1f   0x13
	 {'a' ,'A' ,0x1e,0x01}, //1c    a A    (a)    (A)    0x1e   0x01
	 {'w' ,'W' ,0x11,0x17}, //1d    w W    (w)    (W)    0x11   0x17
	 {'2' ,'@' ,0x79,0   }, //1e    2 @    (2)    (@)    0x79   0
	 {0   ,0   ,0   ,0   }, //1f   //////////////////////////////////
	 {0   ,0   ,0   ,0   }, //20   //////////////////////////////////
	 {'c' ,'C' ,0x2e,0x03}, //21    c C    (c)    (C)    0x2e   0x03

/////////-----------------------SCAN    KEY    NORMAL SHIFT  ALT    CTRL-----

	 {'x' ,'X' ,0x2d,0x18}, //22    x X    (x)    (X)    0x2d   0x18
	 {'d' ,'D' ,0x20,0x04}, //23    d D    (d)    (D)    0x20   0x04
	 {'e' ,'E' ,0x12,0x05}, //24    e E    (e)    (E)    0x12   0x05
	 {'4' ,'$' ,0x7b,0   }, //25    4 $    (4)    ($)    0x7b   0
	 {'3' ,'#' ,0x7a,0   }, //26    3 #    (3)    (#)    0x7a   0
	 {0   ,0   ,0   ,0   }, //27   //////////////////////////////////
	 {0   ,0   ,0   ,0   }, //28   //////////////////////////////////
	 {' ' ,' ' ,0x39,0x73}, //29    SPACE  ( )    ( )    0x39   0x73
	 {'v' ,'V' ,0x2f,0x16}, //2a    v V    (v)    (V)    0x2f   0x16
	 {'f' ,'F' ,0x21,0x06}, //2b    f F    (f)    (F)    0x21   0x06
	 {'t' ,'T' ,0x14,0x14}, //2c    t T    (t)    (T)    0x14   0x14
	 {'r' ,'R' ,0x13,0x12}, //2d    r R    (r)    (R)    0x13   0x12
	 {'5' ,'%' ,0x7c,0   }, //2e    5 %    (5)    (%)    0x7c   0
	 {0   ,0   ,0   ,0   }, //2f   //////////////////////////////////
	 {0   ,0   ,0   ,0   }, //30   //////////////////////////////////
	 {'n' ,'N' ,0x31,0x0e}, //31    n N    (n)    (N)    0x31   0x0e
	 {'b' ,'B' ,0x30,0x02}, //32    b B    (b)    (B)    0x30   0x02

/////////-----------------------SCAN    KEY    NORMAL SHIFT  ALT    CTRL-----

	 {'h' ,'H' ,0x23,0x08}, //33    h H    (h)    (H)    0x23   0x08
	 {'g' ,'G' ,0x22,0x07}, //34    g G    (g)    (G)    0x22   0x07
	 {'y' ,'Y' ,0x15,0x19}, //35    y Y    (y)    (Y)    0x15   0x19
	 {'6' ,'^' ,0x7d,0   }, //36    6 ^    (6)    (^)    0x7d   0
	 {0   ,0   ,0   ,0   }, //37   //////////////////////////////////
	 {0   ,0   ,0   ,0   }, //38   //////////////////////////////////
	 {0   ,0   ,0   ,0   }, //39   //////////////////////////////////
	 {'m' ,'M' ,0x32,0x0d}, //3a    m M    (m)    (M)    0x32   0x0d
	 {'j' ,'J' ,0x24,0x0a}, //3b    j J    (j)    (J)    0x24   0x0a
	 {'u' ,'U' ,0x16,0x15}, //3c    u U    (u)    (U)    0x16   0x15
	 {'7' ,'&' ,0x7e,0   }, //3d    7 &    (7)    (&)    0x7e   0
	 {'8' ,'*' ,0x7f,0   }, //3e    8 *    (8)    (*)    0x7f   0
	 {0   ,0   ,0   ,0   }, //3f   //////////////////////////////////
	 {0   ,0   ,0   ,0   }, //40   //////////////////////////////////
	 {',' ,'<' ,0x82,0   }, //41    , <    (,)    (<)    0x82   0
	 {'k' ,'K' ,0x25,0x0b}, //42    k K    (k)    (K)    0x25   0x0b
	 {'i' ,'I' ,0x17,0x09}, //43    i I    (i)    (I)    0x17   0x09

/////////-----------------------SCAN    KEY    NORMAL SHIFT  ALT    CTRL-----

	 {'o' ,'O' ,0x18,0x0f}, //44    o O    (o)    (O)    0x18   0x0f
	 {'0' ,')' ,0x81,0   }, //45    0 )    (0)    ())    0x81   0
	 {'9' ,'(' ,0x80,0   }, //46    9 (    (9)    (()    0x80   0
	 {0   ,0   ,0   ,0   }, //47   //////////////////////////////////
	 {0   ,0   ,0   ,0   }, //48   //////////////////////////////////
	 {'.' ,'>' ,0x34,0   }, //49    . >    (.)    (>)    0x34   0
	 {'/' ,'?' ,0x35,0x95}, //4a    / ?    (/)    (?)    0x35   0x95
	 {'l' ,'L' ,0x26,0x0c}, //4b    l L    (l)    (L)    0x26   0x0c
	 {';' ,':' ,0x27,0   }, //4c    ; :    (;)    (:)    0x27   0
	 {'p' ,'P' ,0x19,0x10}, //4d    p P    (p)    (P)    0x19   0x10
	 {'-' ,'_' ,0x98,0x8d}, //4e    - _    (-)    (_)    0x98   0x8d
	 {0   ,0   ,0   ,0   }, //4f   //////////////////////////////////
	 {0   ,0   ,0   ,0   }, //50   //////////////////////////////////
	 {0   ,0   ,0   ,0   }, //51   //////////////////////////////////
	 {0x27,'"' ,0x28,0   }, //52    ' "    (')    (")    0x28   0
	 {0   ,0   ,0   ,0   }, //53   //////////////////////////////////
	 {'[' ,'{' ,0x1a,0x1b}, //54    [ {    ([)    ({)    0x1a   0x1b

/////////-----------------------SCAN    KEY    NORMAL SHIFT  ALT    CTRL-----

	 {'=' ,'+' ,0x83,0   }, //55    = +    (=)    (+)    0x83   0
	 {0   ,0   ,0   ,0   }, //56   //////////////////////////////////
	 {0   ,0   ,0   ,0   }, //57   //////////////////////////////////
	 {0   ,0   ,0   ,0   }, //58    CAPS   ---------status----------
				////    LOCK

	 {0   ,0   ,0   ,0   }, //59    SHIFT  ---------status----------
	 {13  ,0   ,0x1c,0   }, //5a    ENTER  13     0      0x1c   0
	 {']' ,'}' ,0x1b,0x1d}, //5b    ] }    (])    (})    0x1b   0x1d
	 {0   ,0   ,0   ,0   }, //5c   //////////////////////////////////
	 {0x5c,'|' ,0x2b,0x1c}, //5d    \ |    (\)    (|)    0x2b   0x1c
	 {0   ,0   ,0   ,0   }, //5e   //////////////////////////////////
	 {0   ,0   ,0   ,0   }, //5f   //////////////////////////////////
	 {0   ,0   ,0   ,0   }, //60   //////////////////////////////////
	 {0   ,0   ,0   ,0   }, //61   //////////////////////////////////
	 {0   ,0   ,0   ,0   }, //62   //////////////////////////////////
	 {0   ,0   ,0   ,0   }, //63   //////////////////////////////////

/////////-----------------------SCAN    KEY    NORMAL SHIFT  ALT    CTRL-----

	 {0   ,0   ,0   ,0   }, //64   //////////////////////////////////
	 {0   ,0   ,0   ,0   }, //65   //////////////////////////////////
	 {29  ,0   ,0x0e,0   }, //66    BACK   29     0      0x0e   0
				////    SPACE

	 {0   ,0   ,0   ,0   }, //67   //////////////////////////////////
	 {0   ,0   ,0   ,0   }, //68   //////////////////////////////////

/////////-----------------------SCAN    KEY    NORMAL NUM    ALT    CTRL-----

	 {0x4f,'1' ,0x9f,0x75}, //69    END    0x4f   (1)    0x9f   0x75
	 {0   ,0   ,0   ,0   }, //6a   //////////////////////////////////
	 {29  ,'4' ,0   ,0   }, //6b    4      29     (4)    0      0
				////    LEFT CURSOR
	 {11  ,'7' ,0x97,0x77}, //6c    7      11     (7)    0x97   0x77
				////    HOME
	 {0   ,0   ,0   ,0   }, //6d   //////////////////////////////////
	 {0   ,0   ,0   ,0   }, //6e   //////////////////////////////////
	 {0   ,0   ,0   ,0   }, //6f   //////////////////////////////////
	 {0x52,'0' ,0xa2,0x92}, //70    0      0x52   (0)    0xa2   0x92
				////    INSERT

	 {0x53,'.' ,0xa3,0x93}, //71    .      0x53   (.)    0xa3   0x93
				////    DELETE
	 {31  ,'2' ,0   ,0   }, //72    2      31     (2)    0      0
				////    DOWN CURSOR
	 {0   ,'5' ,0   ,0x8f}, //73    5      0      (5)    0      0x8f

/////////-----------------------SCAN    KEY    NORMAL NUM    ALT    CTRL-----

	 {28  ,'6' ,0   ,0   }, //74    6      28     (6)    0      0
				////    RIGHT CURSOR
	 {30  ,'8' ,0   ,0   }, //75    8      30     (8)    0      0
				////    UP CURSOR
	 {27  ,0   ,0x01,0   }, //76    ESC    27     0      0x01   0
	 {0   ,0   ,0   ,0   }, //77    NUM    ----------status---------
				////    LOCK

	 {0x85,0x87,0x8b,0x89}, //78    F11    0x85   0x87   0x8b   0x89
	 {'+' ,'+' ,0x4e,0x90}, //79    +      (+)    (+)    0x4e   0x90
	 {0x51,'3' ,0xa1,0x76}, //7a    3      0x51   (3)    0xa1   0x76
				////    PAGE DOWN

	 {'-' ,'-' ,0x4a,0   }, //7b    -      (-)    (-)    0x4a   0
	 {'*' ,'*' ,0x37,0x96}, //7c    *      (*)    (*)    0x37   0x96
	 {0x49,'9' ,0x99,0x84}, //7d    9      0x49   (9)    0x99   0x84
				////    PAGE UP

/////////-----------------------SCAN    KEY    NORMAL SHIFT  ALT    CTRL-----

	 {0   ,0   ,0   ,0   }, //7e    SCROLL ---------status----------
				////    LOCK

	 {0   ,0   ,0   ,0   }, //7f    PRINT SCREEN -------------------
	 {0   ,0   ,0   ,0   }, //80    PAUSE BREAK  -------------------
	 {0   ,0   ,0   ,0   }, //81   //////////////////////////////////
	 {0   ,0   ,0   ,0   }, //82   //////////////////////////////////
	 {0x41,0x5a,0x6e,0x64}, //83    F7     0x41   0x5a   0x6e   0x64
	 };

/////////////////////////READS KEYBOARD DATA LINE////////////////////////////
BOOLEAN get_kbrd_data (void) {
	BOOLEAN state = 0;
	if (inport (PIO1_DATA) & 0x8000)   //poll pio input 31
		state = 1;                          //pc keyboard data line
	return(state);
}

/////////////////////////READS KEYBOARD CLOCK LINE///////////////////////////
BOOLEAN get_kbrd_clock (void) {
	BOOLEAN state = 0;
	if (inport (PIO0_DATA) & 0x0001)   //poll pio input 0
		state = 1;                          //pc keyboard clock line
	return(state);
}

/////////////////////////NON PRECISION DELAY ROUTINE/////////////////////////
void delay_us(UWORD repeat) {
int i,j;
for (j=0;j<=repeat;j++)
	for (i=0;i<=50;i++);
}

/////////////////////////SENDS COMMAND TO KEYBOARD///////////////////////////
void put_kbrd_cmnd(UBYTE command) {

div_t x;
UBYTE parity_count = 0;
UBYTE parity = 0;
UBYTE bit_mask = 0x01;
int i;

PIOconfig (0, PIOCFG_OUTPUT, 0);        //keyboard clock line   initiate
PIOconfig (31, PIOCFG_OUTPUT, 1);        //keyboard data line    command
delay_us(100);                           //                      transmission
PIOconfig (31, PIOCFG_OUTPUT, 0);        //keyboard data line
PIOconfig (0, PIOCFG_INPUT, 1);         //keyboard clock line

while (get_kbrd_clock());                //wait for lo on clock; start bit

for (i=0;i<8;i++) {
	if ((bit_mask & command) != 0) {          //test 'command'
		PIOconfig (31,PIOCFG_OUTPUT, 1);  //1 bit at a time
		++parity_count;                   //count number of 1's
		}                                 //in 'command'
	if ((bit_mask & command) == 0)            //and load data
		PIOconfig (31,PIOCFG_OUTPUT, 0);
	bit_mask = (bit_mask * 2);               //shift bit mask
	while (!get_kbrd_clock());               //wait for hi on clock
	while (get_kbrd_clock());                //wait for lo on clock
}

x = div(parity_count, 2);                //calculate parity
if (x.rem == 0)
	parity = 1;

PIOconfig (31,PIOCFG_OUTPUT, parity);    //clock out parity bit;
while (!get_kbrd_clock());               //wait for hi on clock
while (get_kbrd_clock());                //wait for lo on clock

PIOconfig (31, PIOCFG_INPUT, 1);         //release data line
while (!get_kbrd_clock());               //wait for hi on clock; wait 1 more
while (get_kbrd_clock());                //wait for lo on clock; clock cycle
while (get_kbrd_data());                 //wait for lo on data ; wait for ACK

PIOconfig (0, PIOCFG_OUTPUT, 0);        //keyboard clock line lo puts
					 //keyboard in buffering  mode

}

/////////////////////////READS SCAN CODE FROM KEYBOARD///////////////////////
UBYTE get_scan_code(void) {

UBYTE build_byte = 0;
UBYTE bit_mask = 0x01;
int i;

PIOconfig (0, PIOCFG_INPUT, 1);        //keyboard clock line
PIOconfig (31, PIOCFG_INPUT, 1);        //keyboard data line

while (get_kbrd_clock());                //wait for lo on clock; start bit
for (i=0;i<8;i++) {                      //get 8 data bits, LSB first
	while (!get_kbrd_clock());               //wait for hi on clock
	while (get_kbrd_clock());                //wait for lo on clock
	if (get_kbrd_data())                     //get data bit
	build_byte = (build_byte | bit_mask);    //build scan code
	bit_mask = (bit_mask * 2);               //shift bit mask
	}
while (!get_kbrd_clock());               //wait for hi on clock
while (get_kbrd_clock());                //wait for lo on clock; parity bit
while (!get_kbrd_clock());               //wait for hi on clock
while (get_kbrd_clock());                //wait for lo on clock; stop bit

PIOconfig (0, PIOCFG_OUTPUT, 0);        //keyboard clock line lo puts
					 //keyboard in buffering  mode
return(build_byte);
}

/////////////////////////SETS KEYBOARD LEDS//////////////////////////////////
void set_leds(UBYTE leds) {              //bit 0 controls SCROLL LOCK
					 //bit 1 controls NUM LOCK
UBYTE scan_code;                         //bit 2 controls CAPS LOCK

put_kbrd_cmnd(0xed);                     //send SET STATUS LEDS command
scan_code = get_scan_code();             //wait for ACK from keyboard
if(scan_code == 0xfa);
	put_kbrd_cmnd(leds);             //then send actual led status byte
scan_code = get_scan_code();             //the keyboard sends
}                                        //another ACK here so get it

/////////////////////////PROCESSES SCAN CODES FROM KEYBOARD//////////////////
UBYTE get_scans(void) {

UBYTE temp_scan_code;
int i;

if (at_ascii_code == 0xe0) {                  //if last scan code was ext.
	at_ascii_code = 0;                    //return scan code that
	numpad_flag = 1;                      //came after that
	return (at_scan_code);                //and reset ext. flag
}                                             //also set numpad flag

numpad_flag = 0;                              //reset numpad flag here

temp_scan_code = get_scan_code();             //get first scan code

if (last_scan_code == 0) {                    //is it the first key hit?
	if (temp_scan_code == 0xe0) {         //is it an extended key?
		at_ascii_code = temp_scan_code;       //set ext. flag = e0
		temp_scan_code = get_scan_code();     //get next scan code
		}
	last_scan_code = temp_scan_code;      //then keep track of it
	return(temp_scan_code);               //and return with it
	}

if (temp_scan_code == 0xe1) {                     //is it PAUSE BREAK ?
	for(i=0;i<7;i++)                          //get final 7 scan codes
		temp_scan_code = get_scan_code();
	return (0);                               //return 0
	};

if (temp_scan_code == 0xe0) {                 //is it an extended key?
	at_ascii_code = temp_scan_code;       //set ext. flag = e0
	temp_scan_code = get_scan_code();     //get next scan code
	if (temp_scan_code == 0x12) {         //is it PRINT SCREEN ?
		temp_scan_code = get_scan_code();       //get next two
		temp_scan_code = get_scan_code();       //scan codes
		return (0);                   //return 0
		}
	if (temp_scan_code == 0x5a) {         //is it extended ENTER ?
		at_ascii_code = 0;            //then undo ext. flag
		return (temp_scan_code);      //and return normal ENTER
		}
	if (temp_scan_code == 0x4a) {         //is it extended / ?
		at_ascii_code = 0;            //then undo ext. flag
		return (temp_scan_code);      //and return normal /
		}
	}
						  //is it a function key ?
if ((temp_scan_code >= 0x01) && (temp_scan_code <= 0x0c)
    || (temp_scan_code == 0x83)  || (temp_scan_code == 0x78))
	at_ascii_code = 0xe0;                 //then make it extended

					      //is it a numpad key ?
if ((temp_scan_code >= 0x69) && (temp_scan_code <= 0x75)
    || (temp_scan_code == 0x7a)  || (temp_scan_code == 0x7d))
	if ((at_kbrd_status & 0x10) == 0)     //with NUM LOCK off ?
		at_ascii_code = 0xe0;         //then make it extended

if (temp_scan_code == last_scan_code) {       //is it a typematic repeat?

				 //is it PRINT SCREEN typematic repeat?
	if (at_ascii_code == 0xe0 && temp_scan_code == 0x7c)
		return (0);     //return 0
	else                                      //if not
	return(temp_scan_code);                   //then return with it
	}

if (temp_scan_code == 0xf0) {                 //is it a release code?

	temp_scan_code = get_scan_code();     //then get next scan code

					 //////FLAGS IN STATUS BYTE://////
	if (temp_scan_code == 0x12 || temp_scan_code == 0x59)  //if SHIFT
		at_kbrd_status = (at_kbrd_status ^ 0x01);      //clear shift
	if (temp_scan_code == 0x11)                            //if ALT
		at_kbrd_status = (at_kbrd_status ^ 0x02);      //clear alt
	if (temp_scan_code == 0x14)                            //if CTRL
		at_kbrd_status = (at_kbrd_status ^ 0x04);      //clear ctrl

	return(0);                            //discard it and return 0
	}

if (temp_scan_code != last_scan_code) {       //is it a different key hit?
	last_scan_code = temp_scan_code;      //then keep track of it
	return(temp_scan_code);               //and return with it
	}


return(0);

}

/////////////////////////PROCESSES STATUS KEYS////////////////////////////////
UBYTE at_key_test(void) {

do at_scan_code = get_scans();                 //keep getting scan codes
while (at_scan_code == 0);                     //while they are release codes

					   //////FLAGS IN STATUS BYTE://////
if (at_scan_code == 0x12 || at_scan_code == 0x59) {     //if SHIFT
	at_kbrd_status = (at_kbrd_status | 0x01);       //set shift
	return (0);                                     //return 0
	}
if (at_scan_code == 0x11) {                             //if ALT
	at_kbrd_status = (at_kbrd_status | 0x02);       //set alt
	return (0);                                     //return 0
	}
if (at_scan_code == 0x14) {                             //if CTRL
	at_kbrd_status = (at_kbrd_status | 0x04);       //set ctrl
	return (0);                                     //return 0
	}
if (at_scan_code == 0x58) {                             //if CAPS LOCK
	at_kbrd_status = (at_kbrd_status ^ 0x08);       //toggle caps lock
	at_leds_status = (at_leds_status ^ 0x04);       //toggle led status
	set_leds(at_leds_status);                       //send led command
	return (0);                                     //return 0
	}
if (at_scan_code == 0x77) {                             //if NUM LOCK
	at_kbrd_status = (at_kbrd_status ^ 0x10);       //toggle num lock
	at_leds_status = (at_leds_status ^ 0x02);       //toggle led status
	set_leds(at_leds_status);                       //send led command
	return (0);                                     //return 0
	}
if (at_scan_code == 0x7e) {                             //if SCROLL LOCK
	at_kbrd_status = (at_kbrd_status ^ 0x20);       //toggle scroll lock
	at_leds_status = (at_leds_status ^ 0x01);       //toggle led status
	set_leds(at_leds_status);                       //send led command
	return(0);                                      //return 0
	}

else return(at_scan_code);

}

/////////////////////////CONVERTS SCAN CODES TO ASCII////////////////////////
////////////////////RETURNS 0 IF EXTENDED KEY AND RETURNS////////////////////
///////////////////////EXTENDED KEY VALUE ON NEXT CALL///////////////////////
// A CALL TO at_kbhit() MUST BE DONE BEFORE EACH CALL TO THIS FUNCTION //////
UBYTE at_getch(void) {

//do at_scan_code = at_key_test();               //keep getting scan codes
//while (at_scan_code == 0);                     //while they are release
					       //or status codes

if (at_ascii_code == 0xe0) {                   //if ext. return 0
	return (0);
	}
						 //if numpad or alternate key
if ((at_scan_code >= 0x69) && (at_scan_code <= 0x7d)) { //but not esc. or F11
	if ((at_scan_code != 0x78) && (at_scan_code != 0x76)) {
		if (numpad_flag == 0) {                     //if numpad
			if ((at_kbrd_status & 0x10) != 0) { //if NUM LOCK on
				at_ascii_code = LOOKUP[at_scan_code] [1];
				return(at_ascii_code); //get 2nd table entry
			}
			if ((at_kbrd_status & 0x10) == 0) { //if NUM LOCK off
				at_ascii_code = LOOKUP[at_scan_code] [0];
				return(at_ascii_code); //get 1st table entry
			}
		}
	}
}

if ((at_kbrd_status & 0x02) != 0) {               //if ALT
	at_ascii_code = LOOKUP[at_scan_code] [2]; //get 3rd table entry
	return(at_ascii_code);
	}

if ((at_kbrd_status & 0x04) != 0) {               //if CTRL
	at_ascii_code = LOOKUP[at_scan_code] [3]; //get 4th table entry
	return(at_ascii_code);
	}

// trial lookup to determine if non-alpha
at_ascii_code = LOOKUP[at_scan_code] [0];
// process caps lock condition if alpha
if (isalpha (at_ascii_code)) {
	if ((at_kbrd_status & 0x08) != 0) {                //if CAPS LOCK on
		if ((at_kbrd_status & 0x01) != 0)          //if SHIFT
			at_ascii_code = LOOKUP[at_scan_code] [0]; //get 1st entry
		if ((at_kbrd_status & 0x01) == 0)          //if no SHIFT
			at_ascii_code = LOOKUP[at_scan_code] [1]; //get 2nd entry
		return(at_ascii_code);
	}
}

if ((at_kbrd_status & 0x01) !=0) {                //if SHIFT
	at_ascii_code = LOOKUP[at_scan_code] [1]; //get 2nd table entry
	return(at_ascii_code);
	}

at_ascii_code = LOOKUP[at_scan_code] [0];         //get 1st table entry
return(at_ascii_code);

}


// Works like the kbhit() function in the C library
int at_kbhit (void) {
	at_scan_code = at_key_test();
	if (at_scan_code)
		return -1;
	else
		return 0;
}

int main () {

	UBYTE c;

	printf ("\nDOS Stamp PC keyboard testing, disconnect power to quit...\n");

	while(1) {
		if (at_kbhit()) {
			c = at_getch();
			if (c!=0)
				printf("%c", c);
			if (c==0) {
				at_kbhit();
				c = at_getch();
				printf("\nextended key : %x\n", c);
			}
		}
	}

	return(0);
}

